#!/usr/bin/env python
#
# Copyright (c) 2013, 2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Load output of LTL2DSTAR into a NetworkX DiGraph

The official website of LTL2DSTAR is http://ltl2dstar.de/ , where a
definition of its output format can be found.  Use "-" in place of
FILE to read from stdin.

* Parsing is done by hand; consider changing to use pyparsing.  N.B.,
  the current implementation is lax about the input file, i.e., it
  accepts valid LTL2DSTAR output along with variants.

* The Automaton class is a very light derivative of networkx.DiGraph;
  in other words, if you prefer to only work with a DiGraph, it is
  easy to modify the existing code to do so.  Expect the name to
  change later, especially if (when) it becomes integrated into TuLiP
  and nTLP.

* readdstar.py is a commandline utility.  It expects to be given the
  name of file from which to read (previously recorded) output of
  LTL2DSTAR.  Alternatively, use "-" to read from stdin.  E.g., if
  ltl2dstar and ltl2ba are in the current directory, try

    $ echo 'U a b' | ltl2dstar --ltl2nba=spin:ltl2ba - -|./readdstar.py -

* Each edge (transition) is labeled with two things:
  - "formula" : a disjunctive normal form expression for when the edge
                should be taken; and
  - "subsets of AP": a list of sets of atomic propositions.

  E.g., if AP = {p, q}, and (1,3) is an edge in the Automaton object A,
  then

    A.edge[1][3]["subsets of AP"] = [set([]), set(['p'])]

  means that the transition (1,3) should be taken (assuming current
  execution has led to state 1) if the none of the atomic propositions
  are true (i.e., !p & !q holds), or precisely "p" is true (i.e., p &
  !q holds).


SCL; 2013, 2015.
"""

import sys
import networkx as nx
import matplotlib.pyplot as plt


class AcceptancePair(object):
    def __init__(self, L=None, U=None):
        if L is None:
            self.L = set()
        else:
            self.L = set(L)
        if U is None:
            self.U = set()
        else:
            self.U = set(U)


class Automaton(nx.DiGraph):
    def __init__(self, aut_type=None):
        nx.DiGraph.__init__(self)
        self.aut_type = aut_type

    def __str__(self):
        output = "Type: "
        if self.aut_type == "DRA":
            output += "deterministic Rabin\n"
        elif self.aut_type == "DSA":
            output += "deterministic Streett\n"
        output += "AP = "+str(self.ap)+"\n"
        output += "Transitions:"+"\n"
        output += "\n".join(["\t("+str(u)+", "+str(v)+") :\n\t\tformula: "+str(d["formula"])+"\n\t\tsubsets of AP: "+str(d["subsets of AP"]) for (u,v,d) in self.edges_iter(data=True)])+"\n"
        output += "Acceptance Pairs (each line is of the form (L, U)):"+"\n"
        output += "\n".join(["\t("+str(Fi.L)+", "+str(Fi.U)+")" for Fi in self.F])
        return output


def gen_apformula(AP, intrep):
    """Generate conjunction formula

    >>> gen_apformula(AP=("p", "q"), intrep=2)
    '!p & q'
    """
    return " & ".join([AP[i] if ((intrep >> i) & 1) != 0 else "!"+AP[i] for i in range(len(AP))])

def gen_apsubset(AP, intrep):
    """Generate set of atomic propositions corresponding to integer

    >>> gen_apsubset(AP=("p", "q"), intrep=2)
    set(['q'])
    """
    return set([AP[i] for i in range(len(AP)) if ((intrep >> i) & 1) != 0])


def readdstar(getline):
    """Construct automaton from LTL2DSTAR output.

    getline is any method that can yield successive lines of output
    from LTL2DSTAR.  E.g., a file could be opened and then its
    readline() method passed to readdstar.
    """
    A = None
    aut_type = None
    comments = []

    last_state = -1  # -1 indicates unset
    try:
        while True:
            line = getline()
            if len(line) == 0:
                break  # file.readline() returns empty string at EOF
            parts = line.split()
            if len(parts) == 0:
                continue  # Ignore blank lines

            if not parts[0].endswith(":") and len(parts) == 3:  # id line
                aut_type = parts[0]
                version = parts[1]
                edge_type = parts[2]

                A = Automaton(aut_type=aut_type)
            elif parts[0] == "Comment:":
                comments.append(" ".join(parts[1:]))
            elif parts[0] == "States:":
                num_states = int(parts[1])
            elif parts[0] == "Acceptance-Pairs:":
                num_pairs = int(parts[1])
                A.F = [None for i in range(num_pairs)]
            elif parts[0] == "Start:":
                A.start_state = int(parts[1])
            elif parts[0] == "AP:":
                ap_len = int(parts[1])
                A.ap = tuple([prop.strip("\"").rstrip("\"") for prop in  parts[2:]])
                assert ap_len == len(A.ap)
            elif parts[0] == "State:":
                last_state = int(parts[1])
                apsubset_counter = 0
                A.add_node(last_state)
            elif parts[0] == "Acc-Sig:":
                for accsig in parts[1:]:
                    accsig_index = int(accsig[1:])
                    if A.F[accsig_index] is None:
                        A.F[accsig_index] = AcceptancePair()
                    if accsig[0] == "+":
                        A.F[accsig_index].L.add(last_state)
                    elif accsig[0] == "-":
                        A.F[accsig_index].U.add(last_state)
            elif last_state >= 0 and parts[0] != "---":
                to_state = int(parts[0])
                if not A.has_edge(last_state, to_state):
                    A.add_edge(last_state, to_state)
                    A.edge[last_state][to_state]["formula"] = "("+gen_apformula(A.ap, apsubset_counter)+")"
                    A.edge[last_state][to_state]["subsets of AP"] = [gen_apsubset(A.ap, apsubset_counter)]
                else:
                    A.edge[last_state][to_state]["formula"] += " | ("+gen_apformula(A.ap, apsubset_counter)+")"
                    A.edge[last_state][to_state]["subsets of AP"].append(gen_apsubset(A.ap, apsubset_counter))
                apsubset_counter += 1

    except EOFError:
        pass  # raw_input() throws this at end-of-file

    return A


if __name__ == "__main__":
    if len(sys.argv) < 2 or "-h" in sys.argv:
        print("Usage: "+sys.argv[0]+" FILE")
        exit(1)

    if sys.argv[1] == "-":  # Read from stdin
        getline = raw_input
    else:
        f = open(sys.argv[1], "r")
        getline = f.readline

    print(readdstar(getline) )



#!/usr/bin/env python
"""Tests for the tulip.spec subpackage."""
import logging
logging.basicConfig(level=logging.DEBUG)
logging.getLogger('ltl_parser_log').setLevel(logging.ERROR)
import nose.tools as nt
from tulip.spec import ast, lexyacc
from tulip.spec.parser import parse


def parse_parse_check(formula, expected_length):
    # If expected_length is None, then the formula is malformed, and
    # thus we expect parsing to fail.
    if expected_length is not None:
        assert len(parse(formula)) == expected_length
    else:
        nt.assert_raises(Exception, parse, formula)


def parse_parse_test():
    for (formula, expected_len) in [("G p", 2),
                                    ("p G", None)]:
        yield parse_parse_check, formula, expected_len


def full_name_operators_test():
    formulas = {
        'always eventually p': '( G ( F p ) )',
        'ALwaYs EvenTUAlly(p)': '( G ( F p ) )',
        ('(p and q) UNtIl (q or ((p -> w) and '
         'not (z implies b))) and always next g'):
        ('( ( ( p & q ) U ( q | ( ( p -> w ) & '
         '( ! ( z -> b ) ) ) ) ) & ( G ( X g ) ) )')}

    for f, correct in formulas.iteritems():
        tree = parse(f, full_operators=True)
        # g.write('hehe.png')
        assert tree.flatten() == correct, tree.flatten()


def test_ast_nodes():
    nodes = ast.make_nodes()
    # test Terminal
    t = nodes.Terminal('a')
    assert t.value == 'a'
    assert repr(t) == "Terminal('a')"
    assert str(t) == 'a'
    assert len(t) == 1
    t1 = nodes.Terminal('a')
    assert t == t1
    t2 = nodes.Terminal('b')
    assert t != t2
    assert t.flatten() == 'a'
    # values and operators must be strings
    nt.assert_raises(TypeError, nodes.Terminal, 2)
    nt.assert_raises(TypeError, nodes.Unary, 2, 'v')
    nt.assert_raises(TypeError, nodes.Binary, 2, 'v')
    # test Unary
    u = nodes.Unary('!', t)
    assert u.operator == '!'
    assert u.operands[0] is t
    assert repr(u) == "Unary('!', Terminal('a'))"
    print(u)
    assert str(u) == '(! a)'
    assert len(u) == 2
    assert u.flatten() == '( ! a )'
    # test Binary
    v = nodes.Binary('+', t, t2)
    assert v.operator == '+'
    assert v.operands[0] is t
    assert v.operands[1] is t2
    assert repr(v) == "Binary('+', Terminal('a'), Terminal('b'))"
    assert len(v) == 3
    assert v.flatten() == "( a + b )"
    # different operator map
    opmap = {'!': '!', '+': '+'}
    nodes = ast.make_nodes(opmap)
    assert nodes.Node.opmap is opmap
    assert nodes.Terminal.opmap is opmap
    assert nodes.Unary.opmap is opmap
    assert nodes.Binary.opmap is opmap


def test_fol_nodes():
    nodes = ast.make_fol_nodes()
    # test Var
    v = nodes.Var('a')
    assert v.value == 'a'
    nt.assert_raises(TypeError, nodes.Var, 2)
    # test Bool
    b = nodes.Bool('TRue')
    print(b.value)
    assert b.value == 'True'
    assert b.flatten() == 'True'
    nt.assert_raises(TypeError, nodes.Bool, 2)
    nt.assert_raises(TypeError, nodes.Bool, 'bee')


def test_lex():
    # catch token precedence errors
    # for example if "EVENTUALLY" is defined before "FALSE",
    # then "False" is parsed as "EVENTUALLY", "NAME", where
    # "NAME" is equal to "alse".
    s = 'False'
    lexer = lexyacc.Lexer()
    lexer.lexer.input(s)
    r = list(lexer.lexer)
    assert len(r) == 1
    (tok, ) = r
    assert tok.value == 'False'
    # another case, which needs that "NAME" be defined
    # before "NEXT"
    s = 'X0reach'
    lexer.lexer.input(s)
    r = list(lexer.lexer)
    assert len(r) == 1
    (tok, ) = r
    assert tok.value == 'X0reach'


def lexer_token_precedence_test():
    s = 'False'
    r = parse(s)
    assert isinstance(r, ast.nodes.Bool)
    assert r.value == 'False'
    s = 'a'
    r = parse(s)
    assert isinstance(r, ast.nodes.Var)
    assert r.value == 'a'
    s = 'x = "a"'
    r = parse(s)
    assert isinstance(r, ast.nodes.Binary)
    assert r.operator == '='
    x, a = r.operands
    assert isinstance(x, ast.nodes.Var)
    assert x.value == 'x'
    assert isinstance(a, ast.nodes.Str)
    assert a.value == 'a'
    s = 'y = 1'
    r = parse(s)
    assert isinstance(r, ast.nodes.Binary)
    assert r.operator == '='
    y, n = r.operands
    assert isinstance(y, ast.nodes.Var)
    assert y.value == 'y'
    assert isinstance(n, ast.nodes.Num)
    assert n.value == '1'
    s = '[] a'
    r = parse(s)
    assert isinstance(r, ast.nodes.Unary)
    assert r.operator == 'G'
    assert isinstance(r.operands[0], ast.nodes.Var)
    assert r.operands[0].value == 'a'
    s = 'a U b'
    r = parse(s)
    assert isinstance(r, ast.nodes.Binary)
    assert r.operator == 'U'
    assert isinstance(r.operands[0], ast.nodes.Var)
    assert r.operands[0].value == 'a'
    assert isinstance(r.operands[1], ast.nodes.Var)
    assert r.operands[1].value == 'b'
    s = '( a )'
    r = parse(s)
    assert isinstance(r, ast.nodes.Var)
    s = "(a ' = 1)"
    r = parse(s)
    assert isinstance(r, ast.nodes.Comparator)
    assert r.operator == '='
    x = r.operands[0]
    assert isinstance(x, ast.nodes.Unary)
    assert x.operator == 'X'
    assert isinstance(x.operands[0], ast.nodes.Var)
    assert x.operands[0].value == 'a'
    y = r.operands[1]
    assert isinstance(y, ast.nodes.Num)
    assert y.value == '1'



# Copyright (c) 2012-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""
Interface to the JTLV implementation of GR(1) synthesis

Relevant links:
  - U{JTLV<http://jtlv.ysaar.net/>}
"""
import logging
import os
import re
import subprocess
import tempfile
import warnings
import networkx as nx
from tulip.spec import translation


logger = logging.getLogger(__name__)
DEBUG_SMV_FILE = 'smv.txt'
DEBUG_LTL_FILE = 'ltl.txt'
DEBUG_AUT_FILE = 'aut.txt'


def check_realizable(spec, heap_size='-Xmx128m', priority_kind=-1,
                     init_option=1):
    """Decide realizability of specification defined by given GRSpec object.

    @return: True if realizable, False if not, or an error occurs.
    """
    assert not spec.moore
    assert not spec.plus_one
    fSMV, fLTL, fAUT = create_files(spec)
    realizable = solve_game(spec, fSMV, fLTL, fAUT, heap_size,
                            priority_kind, init_option)
    os.unlink(fSMV)
    os.unlink(fLTL)
    os.unlink(fAUT)
    return realizable


def solve_game(
    spec, fSMV, fLTL, fAUT, heap_size='-Xmx128m',
    priority_kind=3, init_option=1
):
    """Decide realizability of specification.

    Return True if realizable, False if not, or an error occurs.
    Automaton is extracted for file names by fAUT, unless priority_kind == -1


    @type spec: L{GRSpec}
    @param fSMV, fLTL, fAUT: file name for use with JTLV.  This
        enables synthesize() to access them later on.
    @param heap_size: a string that specifies java heap size.

    @param priority_kind: a string of length 3 or an integer that specifies
        the type of priority used in extracting the
        automaton. Possible values of C{priority_kind} are:

            - -1 - No Automaton
            - 3 - 'ZYX'
            - 7 - 'ZXY'
            - 11 - 'YZX'
            - 15 - 'YXZ'
            - 19 - 'XZY'
            - 23 - 'XYZ'

        Here X means that the controller tries to disqualify one of
        the environment assumptions, Y means that the controller tries
        to advance with a finite path to somewhere, and Z means that
        the controller tries to satisfy one of his guarantees.

    @param init_option: an integer that specifies how to handle the
        initial state of the system. Possible values of C{init_option}
        are:

            - 0 - The system has to be able to handle all
                the possible initial system
                states specified on the guarantee side of the specification.

            - 1 (default) - The system can choose its initial state,
                in response to the initial
                environment state. For each initial environment state,
                the resulting automaton contains
                exactly one initial system state,
                starting from which the system can satisfy the specification.

            - 2 - The system can choose its initial state,
                in response to the initial environment state.
                For each initial environment state, the resulting
                automaton contain all the possible initial system states,
                starting from which the system can satisfy the specification.
    """
    priority_kind = get_priority(priority_kind)

    # init_option
    if (isinstance(init_option, int)):
        if (init_option < 0 or init_option > 2):
            warnings.warn("Unknown init_option. Setting it to the default (1)")
            init_option = 1
    else:
        warnings.warn("Unknown init_option. Setting it to the default (1)")
        init_option = 1

    call_jtlv(heap_size, fSMV, fLTL, fAUT, priority_kind, init_option)

    realizable = False

    f = open(fAUT, 'r')
    for line in f:
        if ("Specification is realizable" in line):
            realizable = True
            break
        elif ("Specification is unrealizable" in line):
            realizable = False
            break

    if (realizable and priority_kind > 0):
        print("\nAutomaton successfully synthesized.\n")
    elif (priority_kind > 0):
        print("\nERROR: Specification was unrealizable.\n")

    return realizable


def synthesize(
    spec, heap_size='-Xmx128m', priority_kind=3,
    init_option=1
):
    """Synthesize a strategy satisfying the specification.

    Arguments are described in documentation for L{solve_game}.

    @return: Return strategy as instance of C{networkx.DiGraph}, or a
        list of counter-examples as returned by L{get_counterexamples}.
    """
    assert not spec.moore
    assert not spec.plus_one
    fSMV, fLTL, fAUT = create_files(spec)

    realizable = solve_game(spec, fSMV, fLTL, fAUT, heap_size,
                            priority_kind, init_option)

    if realizable:
        logger.info('loading JTLV output...')

        with open(fAUT, 'r') as f:
            lines = f.readlines()

        logger.info('JTLV returned:\n' + '\n'.join(lines))

        aut = jtlv_output_to_networkx(lines, spec)

        os.unlink(fSMV)
        os.unlink(fLTL)
        os.unlink(fAUT)
        return aut
    else:
        counter_examples = get_counterexamples(fAUT)
        os.unlink(fSMV)
        os.unlink(fLTL)
        os.unlink(fAUT)
        return counter_examples


def create_files(spec):
    """Create temporary files for read/write by JTLV."""
    fSMV = tempfile.NamedTemporaryFile(delete=False, suffix='smv')
    fSMV.write(generate_jtlv_smv(spec))
    fSMV.close()

    fLTL = tempfile.NamedTemporaryFile(delete=False, suffix="ltl")
    fLTL.write(generate_jtlv_ltl(spec))
    fLTL.close()

    fAUT = tempfile.NamedTemporaryFile(delete=False)
    fAUT.close()
    return fSMV.name, fLTL.name, fAUT.name


def get_priority(priority_kind):
    """Validate and convert priority_kind to the corresponding integer.

    @type priority_kind: str or int
    @param priority_kind: a string of length 3 or integer as may be
        used when invoking L{solve_game}.  Check documentation there
        for possible values.

    @rtype: int
    @return: if given priority_kind is permissible, then return
        integer representation of it.  Else, return default ("ZYX").
    """
    if isinstance(priority_kind, str):
        if (priority_kind == 'ZYX'):
            priority_kind = 3
        elif (priority_kind == 'ZXY'):
            priority_kind = 7
        elif (priority_kind == 'YZX'):
            priority_kind = 11
        elif (priority_kind == 'YXZ'):
            priority_kind = 15
        elif (priority_kind == 'XZY'):
            priority_kind = 19
        elif (priority_kind == 'XYZ'):
            priority_kind = 23
        else:
            warnings.warn('Unknown priority_kind.' +
                          'Setting it to the default (ZYX)')
            priority_kind = 3
    elif isinstance(priority_kind, int):
        if (
            priority_kind > 0 and priority_kind != 3 and
            priority_kind != 7 and priority_kind != 11 and
            priority_kind != 15 and priority_kind != 19 and
            priority_kind != 23
        ):
            warnings.warn("Unknown priority_kind." +
                          " Setting it to the default (ZYX)")
            priority_kind = 3
    else:
        warnings.warn("Unknown priority_kind. Setting it to the default (ZYX)")
        priority_kind = 3
    return priority_kind


def call_jtlv(heap_size, fSMV, fLTL, fAUT, priority_kind, init_option):
    """Subprocess calls to JTLV."""
    JTLV_PATH = os.path.abspath(os.path.dirname(__file__))
    JTLV_EXE = 'jtlv_grgame.jar'
    logger.info(
        ('Calling jtlv with arguments:\n'
         '  heap size: {heap}\n'
         '  jtlv path: {path}\n'
         '  priority_kind: {priority}\n').format(
            heap=heap_size,
            path=JTLV_PATH,
            priority=priority_kind
        )
    )
    jtlv_grgame = os.path.join(JTLV_PATH, JTLV_EXE)
    cmd = ['java', heap_size, '-jar', jtlv_grgame, fSMV, fLTL, fAUT,
           str(priority_kind), str(init_option)]
    # debugging log
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('Expected: ' + jtlv_grgame)
        logger.debug(' '.join(cmd))
        # besides dumping to debug logging stream,
        # also copy files to ease manual debugging
        import shutil
        shutil.copyfile(fSMV, DEBUG_SMV_FILE)
        shutil.copyfile(fLTL, DEBUG_LTL_FILE)
        shutil.copyfile(fAUT, DEBUG_AUT_FILE)
    try:
        subprocess.call(cmd)
    except OSError as e:
        if e.errno == os.errno.ENOENT:
            raise Exception('Java not found in path: cannot run jtlv.')
        else:
            raise


def canon_to_jtlv_domain(dom):
    """Convert an LTL or GRSpec variable domain to JTLV style.

    @param dom: a variable domain, as would be provided by the values
    in the output_variables attribute of L{spec.form.LTL} or sys_vars
    of L{spec.form.GRSpec}.

    @rtype: str
    @return: variable domain string, ready for use in an SMV file, as
        expected by the JTLV solver.
    """
    if dom == 'boolean':
        return dom
    elif isinstance(dom, tuple) and len(dom) == 2:
        return '{%s}' % ', '.join([str(i) for i in xrange(dom[0], dom[1]+1)])
    else:
        raise ValueError("Unrecognized domain type: "+str(dom))


def generate_jtlv_smv(spec):
    """Return the SMV module definitions needed by JTLV.

    Raises exception if malformed GRSpec object is detected.

    @type spec: L{GRSpec}

    @rtype: str
    @return: string conforming to SMV file format that JTLV expects.
    """
    # Define env vars
    env = []
    for var, dom in spec.env_vars.iteritems():
        int_dom = translation.convert_domain(dom)

        env.append('\t\t{v} : {c};'.format(
            v=var, c=canon_to_jtlv_domain(int_dom)
        ))

    # Define sys vars
    sys = []
    for var, dom in spec.sys_vars.iteritems():
        int_dom = translation.convert_domain(dom)

        sys.append('\t\t{v} : {c};'.format(
            v=var, c=canon_to_jtlv_domain(int_dom)
        ))

    env_vars = '\n'.join(env)
    sys_vars = '\n'.join(sys)

    smv = smv_template.format(env_vars=env_vars, sys_vars=sys_vars)

    logger.debug('SMV file:\n' + smv)
    return smv


smv_template = '''
MODULE main
    VAR
        e : env();
        s : sys();

MODULE env -- inputs
    VAR
{env_vars}


MODULE sys -- outputs
    VAR
{sys_vars}
'''


def generate_jtlv_ltl(spec):
    """Return LTL specification for JTLV.

    @type spec: L{GRSpec}
    """
    a, g = translation.translate(spec, 'jtlv')

    if not a:
        a = 'TRUE'

    if not g:
        g = 'TRUE'

    ltl = ltl_template.format(a=a, g=g)

    logger.debug('JTLV input formula:\n' + ltl)

    return ltl


ltl_template = '''
LTLSPEC
(
{a}
);

LTLSPEC
(
{g}
);
'''


def jtlv_output_to_networkx(lines, spec):
    """Return a graph after parsing JTLV's output.

    @param lines: as loaded from file
    @type lines: C{list} of C{str}

    @type spec: L{GRSpec}

    @rtype: C{networkx.DiGraph}
    """
    logger.info('parsing jtlv output to create a graph...')

    g = nx.DiGraph()

    varnames = dict(spec.sys_vars)
    varnames.update(spec.env_vars)

    # g.env_vars = spec.env_vars.copy()
    # g.sys_vars = spec.sys_vars.copy()

    for line in lines:
        # parse states
        if line.find('State ') >= 0:
            state_id = re.search('State (\d+)', line)
            state_id = int(state_id.group(1))

            state = dict(re.findall('(\w+):(\w+)', line))
            for var, val in state.iteritems():
                try:
                    state[var] = int(val)
                except:
                    state[var] = val

                if varnames:
                    if var not in varnames:
                        logger.error('Unknown variable "{var}"'.format(
                                     var=var))

            for var, dom in varnames.iteritems():
                # singleton domain ?
                if isinstance(dom, tuple):
                    if dom[0] == dom[1]:
                        state[var] = dom[0]
                if isinstance(dom, list):
                    if len(dom) == 1:
                        state[var] = dom[0]
                # still missing ?
                if var not in state:
                    raise ValueError(
                        'Variable "{var}" not assigned in "{line}"'.format(
                            var=var, line=line))

        # parse transitions
        if line.find('successors') >= 0:
            succ = [int(x) for x in re.findall(' (\d+)', line)]

            g.add_node(state_id, state=state)
            g.add_edges_from([(state_id, v) for v in succ])

    logger.info('done parsing jtlv output.\n')
    return g


def get_counterexamples(aut_file):
    """Return a list of dictionaries, each representing a counter example.

    @param aut_file: a string containing the name of the file
        containing the counter examples generated by JTLV.
    """
    counter_examples = []
    line_found = False
    f = open(aut_file, 'r')
    for line in f:
        if (line.find('The env player can win from states') >= 0):
            line_found = True
            continue
        if (line_found and (len(line) == 0 or line.isspace())):
            line_found = False
        if (line_found):
            counter_ex = dict(re.findall('(\w+):([-+]?\d+)', line))
            for var, val in counter_ex.iteritems():
                counter_ex[var] = int(val)
            counter_examples.append(counter_ex)
            logger.info(counter_ex)
    return counter_examples


def remove_comments(spec):
    """Remove comment lines from string."""
    speclines = spec.split('\n')
    newspec = ''
    for line in speclines:
        if '--' not in line:
            newspec += line + '\n'
    return newspec


def check_vars(varNames):
    """Complain if any variable name is a number or not a string."""
    for item in varNames:
        # Check that the vars are strings
        if type(item) != str:
            logger.error("Prop " + str(item) + " is invalid")
            return False

        # Check that the keys are not numbers
        try:
            int(item)
            float(item)
            logger.errror("Prop " + str(item) + " is invalid")
            return False
        except ValueError:
            continue
    return True


def check_spec(spec, varNames):
    """Verify that all non-operators in "spec" are in the list of vars."""
    # Replace all special characters with whitespace
    special_characters = [
        "next(", "[]", "<>", "->", "&", "|", "!",
        "(", ")", "\n", "<", ">", "<=", ">=", "<->", "\t", "="]
    for word in special_characters:
        spec = spec.replace(word, "")

    # Now, replace all variable names and values with whitespace as well.
    for value in varNames:
        if isinstance(value, (list, tuple)):
            for individual_value in value:
                spec = spec.replace(str(individual_value), "")
        else:
            spec = spec.replace(value, "")

    # Remove all instances of "true" and "false"
    spec = spec.lower()

    spec = spec.replace("true", "")
    spec = spec.replace("false", "")

    # Make sure that the resulting string is empty
    spec = spec.split()

    return not spec



#!/usr/bin/env python
"""Tests for transys.labeled_graphs (part of transys subpackage)"""
from nose.tools import raises, assert_raises
from tulip.transys import labeled_graphs
from tulip.transys.mathset import PowerSet, MathSet
from tulip.transys.transys import FTS


def str2singleton_test():
    assert labeled_graphs.str2singleton("p") == {"p"}
    assert labeled_graphs.str2singleton({"Cal"}) == {"Cal"}


def prepend_with_check(states, prepend_str, expected):
    assert labeled_graphs.prepend_with(states, prepend_str) == expected


def prepend_with_test():
    for (states, prepend_str, expected) in [([0,1], "s", ['s0', 's1']),
                                            ([], "s", []),
                                            ([0], "Cal", ["Cal0"]),
                                            ([0, 1], None, [0, 1])]:
        yield prepend_with_check, states, prepend_str, expected


class States_test():
    def setUp(self):
        self.S = labeled_graphs.States(labeled_graphs.LabeledDiGraph())

    def tearDown(self):
        self.S = None

    def test_contains(self):
        # This also serves as a test for len
        self.S.add(1)
        self.S.add(2)
        assert len(self.S) == 2
        assert (2 in self.S) and (1 in self.S)
        assert 3 not in self.S

    def test_ior(self):
        self.S.add(1)
        other_S = labeled_graphs.States(labeled_graphs.LabeledDiGraph())
        other_S.add(0)
        assert len(self.S) == 1
        assert set([s for s in self.S]) == {1}
        self.S |= other_S
        assert len(self.S) == 2
        assert set([s for s in self.S]) == {1, 0}

    def test_add(self):
        self.S.add(1)
        assert set([s for s in self.S]) == set([1])
        self.S.add(2)
        assert set([s for s in self.S]) == set([1, 2])
        self.S.add("Cal")
        assert set([s for s in self.S]) == set([1, 2, "Cal"])

    def test_add_from(self):
        self.S.add_from(range(3))
        assert len(self.S) == 3
        assert set([s for s in self.S]) == set(range(3))
        self.S.add_from(["Cal", "tech"])
        assert len(self.S) == 5
        assert set([s for s in self.S]) == set(range(3)+["Cal", "tech"])

    def test_remove(self):
        # This also tests remove_from
        self.S.add_from(range(4))
        self.S.remove(1)
        assert set([s for s in self.S]) == set([0, 2, 3])
        self.S.remove_from([0, 3])
        assert len(self.S) == 1 and 2 in self.S

    def test_call(self):
        self.S.add_from([-1, "Cal"])
        S_imm_dat = self.S(data=True)
        assert (len(S_imm_dat) == 2) and ((-1, dict()) in S_imm_dat) and \
            (("Cal", dict()) in S_imm_dat)

    def test_postpre(self):
        self.S.add_from(range(5))
        self.S.graph.add_edges_from([(0, 1), (0, 2), (1, 3), (3, 4)])
        assert self.S.post(0) == {1, 2}
        assert self.S.post([0, 1]) == {1, 2, 3}
        assert self.S.pre(4) == {3}
        assert self.S.pre([1, 2, 4]) == {0, 3}

    def test_is_terminal(self):
        self.S.add_from([0, 1])
        self.S.graph.add_edge(0, 1)
        assert not self.S.is_terminal(0)
        assert self.S.is_terminal(1)


class Transitions_test:
    def setUp(self):
        G = labeled_graphs.LabeledDiGraph()
        self.T = labeled_graphs.Transitions(G)
        G.transitions = self.T
        self.T.graph.states.add_from([1, 2, 3, 4, 5])

    def tearDown(self):
        self.T = None

    def test_len(self):
        assert len(self.T) == 0
        self.T.add(1, 2)
        assert len(self.T) == 1
        self.T.add(2, 3)
        assert len(self.T) == 2

        # Transitions should be unaffected by new states
        self.T.graph.states.add_from([10])
        assert len(self.T) == 2

    @raises(Exception)
    def test_missing_states(self):
        self.T.add(10, 11, check=True)

    def test_add_from(self):
        self.T.add(1, 4)
        assert len(self.T) == 1 and set([t for t in self.T()]) == set([(1, 4)])
        assert_raises(Exception, self.T.add, 1, 4)
        assert len(self.T) == 1  # Edge already exists, so not added

        self.T.add_from([(5, 2), (4, 3)])
        assert len(self.T) == 3
        assert set([t for t in self.T()]) == set([(1, 4), (5, 2), (4, 3)])

    def test_add_comb(self):
        self.T.add_comb([1, 2], [3, 4])
        assert len(self.T) == 4 and set([t for t in self.T()]) == set([(1, 3),
                                                                       (2, 3),
                                                                       (1, 4),
                                                                       (2, 4)])

    def test_remove(self):
        # This also tests remove_from
        self.T.add_from([(1, 2), (1, 3), (4, 3), (3, 2)], check=False)
        assert len(self.T) == 4
        self.T.remove(1, 2)
        assert len(self.T) == 3
        assert set([t for t in self.T()]) == set([(1, 3), (4, 3), (3, 2)])
        self.T.remove_from([(1, 2), (4, 3), (3, 2)])
        assert set([t for t in self.T()]) == set([(1, 3)])


class States_labeling_test:
    def setUp(self):
        node_label_def = [{
            'name': 'ap',
            'values': PowerSet({'p', 'q', 'r', 'x', 'a', 'b'})}]
        G = labeled_graphs.LabeledDiGraph(node_label_def)
        self.S_ap = labeled_graphs.States(G)
        G.states = self.S_ap

    def tearDown(self):
        self.S_ap = None

    @raises(Exception)
    def test_add_untyped_keys(self):
        self.S_ap.add(1, foo=MathSet(['p']), check=True)

    def test_add(self):
        self.S_ap.add(1, ap={'p'} )
        assert len(self.S_ap) == 1
        self.S_ap.add(2, ap={'p'} )
        assert len(self.S_ap) == 2
        self.S_ap.add(3, ap={'q', 'r'} )
        assert len(self.S_ap) == 3

        assert self.S_ap[1] == {'ap': {'p'} }
        assert self.S_ap[3] == {'ap': {'q', 'r'} }

        nodes = {u for u,l in self.S_ap.find(
            with_attr_dict={'ap': {'p'} })}
        assert nodes == set([1, 2])

    def test_add_from(self):
        self.S_ap.add_from([(0, {'ap':{'p'}}), (1, {'ap':{'q'} })])
        assert len(self.S_ap) == 2
        current_labels = [l["ap"] for (s,l) in self.S_ap(data=True)]

        assert len(current_labels) == 2 and \
            MathSet(current_labels) == MathSet([{'p'}, {'q'}])

        self.S_ap.add_from([(10, {'ap':{'r'} }),
                                      (11, {'ap':{'x'} })],
                                      check=False)
        assert len(self.S_ap) == 4
        self.S_ap.add_from([(10, {'ap':{'a'} }),
                                      (11, {'ap':{'b'} })])
        assert len(self.S_ap) == 4
        current_labels = [l["ap"] for (s,l) in self.S_ap(data=True)]
        assert len(current_labels) == 4 and \
            MathSet(current_labels)== MathSet([{'p'}, {'q'}, {'a'}, {'b'}])
        assert MathSet([l["ap"] for (s,l) in self.S_ap(data=True)]) == MathSet(current_labels)

    def test_find(self):
        state_list = ["state"+str(i) for i in range(4)]
        state_list = zip(state_list,
                         [{"ap": L} for L in [{'p'}, {'q'}, {'p'}, {'q'}]])
        self.S_ap.add_from(state_list, check=False)
        result = self.S_ap.find("state1")
        assert len(result) == 1 and result[0] == ("state1", {"ap": set(['q'])})

        result = self.S_ap.find(["state1", "state0"])
        assert len(result) == 2 and \
           ("state1", {"ap": set(['q'])}) in result and \
           ("state0", {"ap": set(['p'])}) in result

        result = self.S_ap.find(with_attr_dict={"ap": {'p'}})
        print result
        assert len(result) == 2 and \
           set([s for (s, l) in result]) == set(["state0", "state2"])

        same_result = self.S_ap.find(ap={'p'})
        assert(same_result == result)


class LabeledDiGraph_test():
    def setUp(self):
        p = PowerSet({1, 2})
        node_labeling = [
            {
                'name': 'month',
                'values': ['Jan', 'Feb']
            },
            {
                'name': 'day',
                'values': ['Mon', 'Tue']
            },
            {
                'name': 'comb',
                'values': p,
                'setter': p.math_set
            }
        ]
        edge_labeling = node_labeling
        G = labeled_graphs.LabeledDiGraph(node_labeling, edge_labeling)

        G.states.add_from({1, 2})
        G.transitions.add(1, 2, month='Jan', day='Mon')

        assert_raises(Exception, G.transitions.add,
                      1, 2, {'month': 'Jan', 'day': 'abc'})

        # note how untyped keys can be set directly via assignment,
        # whereas check=False is needed for G.add_node
        G.node[1]['mont'] = 'Feb'
        assert(G.node[1] == {'mont':'Feb'})

        G[1][2][0]['day'] = 'Tue'
        assert(G[1][2][0] == {'month':'Jan', 'day':'Tue'})

        self.G = G

    @raises(AttributeError)
    def test_add_edge_only_typed(self):
        """check that untyped attribute keys are caught"""
        self.G.add_edge(1, 2, mo='Jan')

    def test_add_edge_untyped(self):
        """the untyped attribute key 'mo' should be allowed,
        because check=False
        """
        self.G.add_edge(1, 2, mo='Jan', check=False)
        assert(self.G[1][2][1] == {'mo':'Jan'})

    @raises(ValueError)
    def test_add_edge_illegal_value(self):
        self.G.add_edge(1, 2, month='haha')

    @raises(ValueError)
    def test_node_subscript_assign_illegal_value(self):
        self.G.node[1]['month'] = 'abc'

    @raises(ValueError)
    def test_edge_subscript_assign_illegal_value(self):
        self.G[1][2][0]['day'] = 'abc'


def open_fts_multiple_env_actions_test():
    env_modes = MathSet({'up', 'down'})
    env_choice = MathSet({'left', 'right'})

    env_actions = [
        {
            'name': 'env_modes',
            'values': env_modes,
            'setter': True},
        {
            'name': 'env_choices',
            'values': env_choice}]
    ts = FTS(env_actions)
    assert(ts.env_modes is env_modes)
    assert(not hasattr(ts, 'env_choices') )
    assert(ts.sys_actions == MathSet() )



def test_remove_deadends():
    g = labeled_graphs.LabeledDiGraph()

    # cycle
    n = 5
    g.add_nodes_from(range(n))
    for i in range(n):
        j = (i + 1) % n
        g.add_edge(i, j)

    g.remove_deadends()
    assert(len(g) == n)

    # line + cycle
    g.add_nodes_from(range(n, 2*n))
    for i in xrange(n, 2*n-1):
        g.add_edge(i, i+1)
    assert(len(g) == 2*n)

    g.remove_deadends()
    assert(len(g) == n)

    # line + self-loop
    g.remove_edge(4, 0)
    g.add_edge(0, 0)

    g.remove_deadends()
    assert(len(g) == 1)



# Copyright (c) 2013-2015 by California Institute of Technology
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the California Institute of Technology nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CALTECH
# OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
"""Mathematical Sets and Power Sets"""
import logging
import warnings
from itertools import chain, combinations
from collections import Iterable, Hashable, Container
from pprint import pformat
from random import randint


logger = logging.getLogger(__name__)


def compare_lists(list1, list2):
    """Compare list contents, ignoring ordering.

    Hashability of elements not assumed, incurring O(N**2)

    See Also
    ========
    L{MathSet}

    @type list1: list
    @type list2: list

    @return: True if a bijection exists between the lists.
        Note that this takes into account multiplicity of elements.
    @rtype: bool
    """
    if not isinstance(list1, list):
        raise TypeError('Not a list, instead list1:\n\t' + str(list1))

    if not isinstance(list2, list):
        raise TypeError('Not a list, instead list2:\n\t' + str(list2))
    dummy_list = list(list1)
    same_lists = True
    for item in list2:
        try:
            dummy_list.remove(item)
        except:
            # unique element not found to be rm'd
            same_lists = False
            break
    # anything remaining ?
    same_lists = same_lists and not bool(dummy_list)
    return same_lists


class MathSet(object):
    """Mathematical set, allows unhashable elements.

    Examples
    ========
    >>> s = MathSet(['a', 1, [1,2], {'a', 'b'} ] )

    Then print(s) shows how the elements were separately stored
    in a set and list, to optimize contains operations:

    >>> print(s)
    MathSet(['a', 1, [1, 2], set(['a', 'b'])])

    Set operations similar to the builtin type are supported:

    >>> p = MathSet()
    >>> p.add(1)
    >>> p |= [1, 2]
    >>> p |= {3, 4}
    >>> p |= [[1, 2], '5', {'a':1} ]
    >>> p.add_from([5, 6, '7', {8, 9} ] )
    >>> p.remove(1)
    >>> p
    MathSet([2, 3, 4, 5, 6, '5', '7', [1, 2], {'a': 1}, set([8, 9])])

    See Also
    ========
    L{SubSet}, L{PowerSet}, set
    """

    def __init__(self, iterable=[]):
        """Initialize by adding elements from iterable.

        Example
        =======
        >>> s = MathSet([1, 2, 'a', {3, 4} ] )

        @param iterable: iterable from which to initialize the set S
            which underlies the PowerSet 2^S
        @type iterable: iterable, any element types allowed
        """
        self._delete_all()
        self.add_from(iterable)

    def __repr__(self):
        return 'MathSet(' + pformat(list(self._set) + self._list) + ')'

    def _debug_repr(self):
        set_str = ', '.join([repr(i) for i in self._set])
        return 'MathSet({' + set_str + '} +' + str(self._list) + ')'

    def __or__(self, other):
        """Union with another mathematical set.

        See Also
        ========
        L{__ior__}

        @param other: any other mathematical set.
        @type other: iterable, elements not restricted to hashable

        @return: self | iterable
        @rtype: MathSet
        """
        s = MathSet(self)
        s.add_from(other)
        return s

    def __mul__(self, other):
        """Return the Cartesian product with another C{MathSet}.

        Example
        =======
        >>> a = MathSet([1, 2] )
        >>> b = MathSet([3, 4] )
        >>> c = a *b
        >>> print(type(c) )

        >>> print(c)


        If we prefer a CartesianProduct returned instead:
        >>> c = a.cartesian(b)

        See Also
        ========
        L{cartesian}

        @param other: set with which to take Cartesian product
        @type other: MathSet

        @return: Cartesian product of C{self} with C{other}.
        @rtype: C{MathSet} (explicit construction)
        """
        cartesian = [(x, y) for x in self for y in other]
        return MathSet(cartesian)

    def __ior__(self, iterable):
        """Union with of MathSet with iterable.

        Example
        =======
        >>> s = MathSet([1, 2] )
        >>> s |= [3, 4] # much cleaner & familiar
        >>> print(s)
        set([1, 2, 3, 4]) U []

        See Also
        ========
        L{__or__}

        @param iterable: any mathematical set.
        @type iterable: iterable, elements not restricted to hashable

        @return: self | iterable
        @rtype: MathSet
        """
        self.add_from(iterable)
        return self

    def __sub__(self, rm_items):
        s = MathSet(self)
        print('s = ' + str(s))
        print(rm_items)
        for item in rm_items:
            if item in s:
                print('Removing...: ' + str(item))
                s.remove(item)
        return s

    def __isub__(self, rm_items):
        """Delete multiple elements."""
        for item in rm_items:
            if item in self:
                self.remove(item)
        return self

    def __eq__(self, other):
        if not isinstance(other, MathSet):
            raise TypeError(
                'For now comparison only to another MathSet.\n'
                'Got:\n\t' + str(other) + '\n of type: ' +
                str(type(other)) + ', instead.')
        same_lists = compare_lists(self._list, other._list)
        return (self._set == other._set) and same_lists

    def __contains__(self, item):
        if isinstance(item, Hashable):
            try:
                return item in self._set
            except:
                logger.error('UnHashable items within Hashable.')
        return item in self._list

    def __iter__(self):
        return iter(self._list + list(self._set))

    def __len__(self):
        """Number of elements in set."""
        return len(self._set) + len(self._list)

    def _filter_hashables(self, iterable):
        return filter(lambda x: isinstance(x, Hashable), iterable)

    def _filter_unhashables(self, iterable):
        return list(filter(lambda x: not isinstance(x, Hashable), iterable))

    def _delete_all(self):
        self._set = set()
        self._list = list()

    def add(self, item):
        """Add element to mathematical set.

        Example
        =======
        >>> s = MathSet()
        >>> s.add(1)
        set([1]) U []

        See Also
        ========
        L{add_from}, L{__ior__}, L{remove}

        @param item: the new set element
        @type item: anything, if hashable it is stored in a Python set,
            otherwise stored in a list.
        """
        if isinstance(item, Hashable):
            try:
                self._set.add(item)
                return
            except TypeError:
                logger.error('UnHashable items within Hashable.')
        if item not in self._list:
            self._list.append(item)
        else:
            logger.warn('item already in MathSet.')

    def add_from(self, iterable):
        """Add multiple elements to mathematical set.

        Equivalent to |=

        Example
        =======
        >>> s = MathSet()
        >>> s.add_from([1, 2, {3} ] )

        is equivalent to:

        >>> s = MathSet()
        >>> s |= [1, 2, {3} ]

        See Also
        ========
        L{add}, L{__ior__}, L{remove}

        @param iterable: new MathSet elements
        @type iterable: iterable containing (possibly not hashable) elements
        """
        if not isinstance(iterable, Iterable):
            raise TypeError(
                'Can only add elements to MathSet from Iterable.\n'
                'Got:\n\t' + str(iterable) + '\n instead.')
        if isinstance(iterable, MathSet):
            self._set |= set(iterable._set)
            self._list = list(unique(self._list + iterable._list))
            return
        # speed up
        if isinstance(iterable, set):
            self._set |= iterable
            return
        # filter to optimize storage
        try:
            self._set |= set(self._filter_hashables(iterable))
            self._list = list(unique(
                self._list + self._filter_unhashables(iterable)))
            return
        except:
            # ...if contents of elements in iterable are mutable
            self._list = list(unique(self._list + list(iterable)))

    def remove(self, item):
        """Remove existing element from mathematical set.

        Example
        =======
        >>> p = MathSet([1, 2] )
        >>> p.remove(1)
        >>> p
        set([2]) U []

        See Also
        ========
        L{add}, L{add_from}, L{__or__}

        @param item: An item already in the set.
            For adding items, see add.
        """
        if item not in self:
            warnings.warn(
                'Set element not in set S.\n'
                'Maybe you targeted another element for removal ?')
        if isinstance(item, Hashable):
            try:
                self._set.remove(item)
                return
            except:
                logger.debug('item: ' + str(item) + ', contains unhashables.')
        self._list.remove(item)

    def pop(self):
        """Remove and return random MathSet element.

        Raises KeyError if MathSet is empty.
        """
        if not self:
            raise KeyError('Nothing to pop: MathSet is empty.')
        if self._set and self._list:
            if randint(0, 1):
                return self._set.pop()
            else:
                return self._list.pop()
        elif self._set and not self._list:
            return self._set.pop()
        elif self._list and not self._set:
            return self._list.pop()
        else:
            raise Exception('Bug in empty MathSet: not self above' +
                            'should not reaching this point.')

    def intersection(self, iterable):
        """Return intersection with iterable.

        @param iterable: find common elements with C{self}
        @type iterable: C{Iterable}

        @return: intersection of C{self} with C{iterable}
        @rtype: C{MathSet}
        """
        s = MathSet()
        for item in iterable:
            print(item)
            if item in self:
                print('Adding...\n')
                s.add(item)
        return s

    def intersects(self, iterable):
        """Check intersection with iterable.

        Checks the existence of common elements with iterable.

        >>> s = MathSet()
        >>> s.add(1)
        >>> r = [1,2]
        >>> s.intersects(r)
        True

        @param iterable: with which to check intersection
        @type iterable: C{Iterable}

        @return: C{True} if C{self} has common element with C{iterable}.
            Otherwise C{False}.
        @rtype: C{bool}
        """
        for item in iterable:
            if item in self:
                return True
        return False


class SubSet(MathSet):
    """Subset of selected MathSet, or other Iterable.

    Prior to adding new elements,
    it checks that they are in its superset.

    Example
    =======
    >>> superset = [1, 2]
    >>> s = SubSet(superset)
    >>> s |= [1, 2]
    >>> print(s)
    SubSet([[1, 2]])
    >>> s.add(3)
    raises exception because 3 \\notin [1,2]

    See Also
    ========
    L{MathSet}, L{PowerSet}
    """

    def __init__(self, superset, iterable=None):
        """Define the superset of this set.

        @param superset: This SubSet checked vs C{superset}
        @type superset: Iterable

        @param iterable: elements to add to subset
        @type iterable: Iterable
        """
        self._superset = superset
        super(SubSet, self).__init__([])
        if not isinstance(superset, Container):
            raise TypeError('superset must be Iterable,\n'
                            'Got instead:\n\t' + str(superset))

    def __repr__(self):
        return 'SubSet(' + pformat(list(self._set) + self._list) + ')'

    def _debug_repr(self):
        set_str = ', '.join([repr(i) for i in self._set])
        return 'SubSet({' + set_str + '} +' + str(self._list) + ')'

    @property
    def superset(self):
        return self._superset

    def add(self, new_element):
        """Add state to subset.

        Extends MathSet.add with subset relation checking.

        Example
        =======
        C{new_initial_state} should already be a state.
        First use states.add to include it in set of states,
        then states.add_initial.

        See Also
        ========
        L{MathSet.add}
        """
        if new_element not in self._superset:
            raise Exception(
                'New element state \\notin superset.\n'
                'Add it first to states using e.g. sys.states.add()\n'
                'FYI: new element:\n\t' + str(new_element) + '\n'
                'and superset:\n\t' + str(self._superset))
        super(SubSet, self).add(new_element)

    def add_from(self, new_elements):
        """Add multiple new elements to subset.

        Extends MathSet.add_from with subset relation checking.

        Note
        ====
        It would be sufficient to extend only .add provided
        MathSet.add_from called .add iteratively.
        However MathSet.add_from filters states, which is
        arguably more efficient. So both .add and .add_from
        need to be extended here.

        See Also
        ========
        L{add}, L{__ior__}
        """
        if not is_subset(new_elements, self._superset):
            raise Exception('All new_elements:\n\t' + str(new_elements) +
                            '\nshould already be \\in ' +
                            'self.superset = ' + str(self._superset))
        super(SubSet, self).add_from(new_elements)


class CartesianProduct(object):
    """List of MathSets, with Cartesian semantics."""

    def __init__(self):
        self.mathsets = []

    def __contains__(self, element):
        # TODO check ordered
        if not isinstance(element, Iterable):
            raise TypeError(
                'Argument element must be Iterable, otherwise cannot '
                'recover which item in it belongs to which set in the '
                'Cartesian product.')
        for idx, item in enumerate(element):
            if item not in self.mathsets[idx]:
                return False
        return True

    def __mul__(self, mathsets):
        """Multiply Cartesian products."""
        if not isinstance(mathsets, list):
            raise TypeError('mathsets given must be a list of MathSet.')

    def add(self, mathset):
        self.mathsets += [mathset]

    def add_from(self, mathsets):
        self.mathsets += mathsets

    def remove(self, mathset):
        self.mathsets.remove(mathset)

    def remove_from(self, mathsets):
        for mathset in mathsets:
            self.remove(mathset)


def unique(iterable):
    """Return unique elements.

    Note
    ====
    Always returning a list for consistency was tempting,
    however this defeats the purpose of creating this function
    to achieve brevity elsewhere in the code.

    @return: iterable with duplicates removed, as C{set} if possible.
    @rtype:
        - If all items in C{iterable} are hashable,
            then returns C{set}.
        - If iterable contains unhashable item,
            then returns C{list} of unique elements.
    """
    # hashable items ?
    try:
        unique_items = set(iterable)
    except:
        unique_items = []
        for item in iterable:
            if item not in unique_items:
                unique_items.append(item)
    return unique_items


def contains_multiple(iterable):
    """Does iterable contain any item multiple times ?"""
    return len(iterable) != len(unique(iterable))


def is_subset(small_iterable, big_iterable):
    """Comparison for handling list <= set, and lists with unhashable items.
    """
    # asserts removed when compiling with optimization on...
    # it would have been elegant to use instead:
    #   assert(isinstance(big_iterable, Iterable))
    # since the error msg is succintly stated by the assert itself
    if not isinstance(big_iterable, (Iterable, Container)):
        raise TypeError('big_iterable must be either Iterable or Container, '
                        'otherwise subset relation undefined.\n'
                        'Got:\n\t' + str(big_iterable) + '\ninstead.')
    if not isinstance(small_iterable, Iterable):
        raise TypeError('small_iterable must be Iterable, '
                        'otherwise subset relation undefined.\n'
                        'Got:\n\t' + str(small_iterable) + '\ninstead.')
    # nxor
    if isinstance(small_iterable, str) != isinstance(big_iterable, str):
        raise TypeError('Either both or none of small_iterable, '
                        'big_iterable should be strings.\n'
                        'Otherwise subset relation between string '
                        'and non-string may introduce bugs.\nGot:\n\t' +
                        str(small_iterable) + ',\t' + str(big_iterable) +
                        '\ninstead.')
    try:
        # first, avoid object duplication
        if not isinstance(small_iterable, set):
            small_iterable = set(small_iterable)
        if not isinstance(big_iterable, set):
            big_iterable = set(big_iterable)
        return small_iterable <= big_iterable
    except TypeError:
        # not all items hashable...
        try:
            # list to avoid: unhashable \in set ? => error
            if not isinstance(big_iterable, list):
                # avoid object duplication
                big_iterable = list(big_iterable)
        except:
            logger.error('Could not convert big_iterable to list.')

        for item in small_iterable:
            if item not in big_iterable:
                return False
        return True
    except:
        raise Exception('Failed to compare iterables.')


def powerset(iterable):
    """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)

    From http://docs.python.org/2/library/itertools.html,
    also in https://pypi.python.org/pypi/more-itertools
    """
    s = list(iterable)
    return chain.from_iterable(combinations(s, r)
                               for r in xrange(len(s) + 1))


class PowerSet(object):
    """Efficiently store power set of a mathematical set.

    Set here isn't necessarily a Python set,
    i.e., it may comprise of unhashable elements.

    Example
    =======
    Specify the mathematical set S underlying the PowerSet.

    >>> S = [[1, 2], '3', {'a':1}, 1]
    >>> p = PowerSet(S)

    >>> q = PowerSet()
    >>> q.math_set = S

    Add new element to underlying set S.

    >>> p.math_set.add({3: 'a'} )

    Add multiple new elements to underlying set S.

    >>> p.math_set.add_from({3, 'a'} )
    >>> p.math_set |= [1,2]

    Remove existing element from set S.

    >>> p.remove(1)

    See Also
    ========
    L{MathSet}, L{SubSet}, L{is_subset}

    @param iterable: mathematical set S of elements, on which this 2^S defined.
    @type iterable: iterable container
    """

    def __init__(self, iterable=None):
        """Create new PowerSet over elements contained in S = C{iterable}.

        This powerset is 2^S.

        @param iterable: contains elements of set S underlying the PowerSet.
        @type iterable: iterable of elements which can be hashable or not.
        """
        if iterable is None:
            iterable = []
        self.math_set = MathSet(iterable)

    def __get__(self, instance, value):
        return self()

    def __repr__(self):
        return 'PowerSet(' + str(self.math_set) + ' )'

    def __contains__(self, item):
        """Is item \\in 2^iterable = this powerset(iterable)."""
        if not isinstance(item, Iterable):
            raise Exception('Not iterable:\n\t' + str(item) + ',\n'
                            'this is a powerset, so it contains (math) sets.')

        return is_subset(item, self.math_set)

    def __iter__(self):
        return powerset(self.math_set)

    def __len__(self):
        return 2 ** len(self.math_set)

    def __add__(self, other):
        if not isinstance(other, PowerSet):
            raise TypeError('Addition defined only between PowerSets.\n'
                            'Got instead:\n\t other = ' + str(other))
        list1 = self.math_set
        list2 = other.math_set
        union = list1 | list2
        return PowerSet(union)

    def __eq__(self, other):
        if not isinstance(other, PowerSet):
            raise TypeError('Can only compare to another PowerSet.')

        return other.math_set == self.math_set

    def __setattr__(self, name, value):
        if name is 'math_set' and not isinstance(value, MathSet):
            msg = (
                'PowerSet.math_set must be of class MathSet.\n'
                'Got instead:\n\t' + str(value) +
                '\nof class:\nt\t' + str(type(value)))
            raise Exception(msg)
        object.__setattr__(self, name, value)


class TypedDict(dict):
    """dict subclass where values can be constrained by key.

    For each key, a domain can optionally be defined,
    which restricts the admissible values that can be
    paired with that key.

    Example
    =======

    >>> d = TypedDict()
    >>> allowed_values = {'name': {'Maria', 'John'},
                          'age': range(122)}
    >>> default_values = {'name': 'Maria',
                          'age': 30}
    >>> d.set_types(allowed_types)
    >>> d.update(default_values)
    """
    # credits for debugging this go here:
    #   http://stackoverflow.com/questions/2060972/

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)
        self.allowed_values = dict()

    def __setitem__(self, i, y):
        """Raise ValueError if value y not allowed for key i."""
        valid_y = True
        if hasattr(self, 'allowed_values') and i in self.allowed_values:
            valid_y = False
            if self.allowed_values[i] is None:
                valid_y = True
            else:
                try:
                    if y in self.allowed_values[i]:
                        valid_y = True
                except:
                    valid_y = False
        if not valid_y:
            msg = (
                'key: ' + str(i) + ', cannot be'
                ' assigned value: ' + str(y) + '\n'
                'Admissible values are:\n\t'
                + str(self.allowed_values[i]))
            raise ValueError(msg)
        super(TypedDict, self).__setitem__(i, y)

    def __str__(self):
        return 'TypedDict(' + dict.__str__(self) + ')'

    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                raise TypeError("update expected at most 1 arguments, "
                                "got %d" % len(args))
            other = dict(args[0])
            for key in other:
                self[key] = other[key]
        for key in kwargs:
            self[key] = kwargs[key]

    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        return self[key]

    def set_types(self, allowed_values):
        """Restrict values the key can be paired with.

        @param allowed_values: dict of the form::

                {key : values}

            C{values} must implement C{__contains__}
            to enable checking validity of values.

            If C{values} is C{None},
            then any value is allowed.
        """
        self.allowed_values = allowed_values

    def is_consistent(self):
        """Check if typed keys have consistent values.

        Use case: changing the object that allowed_values
        points to can invalidate the assigned values.

        @rtype: bool
        """
        for k, v in self:
            if k in self.allowed_values:
                if v in self.allowed_values[k]:
                    return False
        return True
